# SPDX-FileCopyrightText: 2022 German Aerospace Center <amiris@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

import os
from pathlib import Path

from amirispy.source.fameio_calls import compile_output, call_amiris, compile_input, determine_all_paths

from amirispy.source.cli import RunOptions
from amirispy.source.files import check_if_write_access
from amirispy.source.logs import log_error_and_raise
from amirispy.source.util import check_java

_ERR_NOT_A_FILE = "Specified path '{}' is no file."


def run_amiris(options: dict) -> None:
    """
    Compile scenario to protobuf using fameio.scripts.make_config,
    execute AMIRIS, and extract results using fameio.scripts.convert_results

    Args:
        options: dictionary of command line instructions

    Returns:
        None
    """
    check_java(skip=options[RunOptions.NO_CHECKS])
    origin_wd = Path.cwd()
    check_if_write_access(origin_wd)

    path_to_scenario: Path = options[RunOptions.SCENARIO]
    if not path_to_scenario.is_file():
        log_error_and_raise(ValueError(_ERR_NOT_A_FILE.format(path_to_scenario)))

    paths = determine_all_paths(path_to_scenario, origin_wd, options, batch=False)
    os.chdir(paths["SCENARIO_DIRECTORY"])
    compile_input(options, paths)
    os.chdir(origin_wd)
    call_amiris(paths)
    compile_output(options, paths)



# SPDX-FileCopyrightText: 2025 German Aerospace Center <amiris@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

import argparse
from enum import Enum, auto
from pathlib import Path
from typing import Any, Dict, Tuple, Optional, List

from amirispy.scripts.subcommands.download import DownloadMode
from amirispy.source.logs import LogLevels


class CommandLineError(Exception):
    """An error that occurred while parsing the command line options"""


AMIRIS_PARSER = "Command-line interface to the electricity market model AMIRIS"
AMIRIS_LOG_FILE_HELP = "Provide logging file (default: None)"
AMIRIS_LOG_LEVEL_HELP = f"Choose logging level (default: {LogLevels.ERROR.name})"
AMIRIS_COMMAND_HELP = "Choose one of the following commands:"

INSTALL_DEPRECATION = "'install' command is deprecated: use 'download' instead."
DOWNLOAD_HELP = "Downloads and extracts latest open access AMIRIS instance"
DOWNLOAD_URL_MODEL_HELP = "URL to download AMIRIS model from (default: latest AMIRIS artifact)"
DOWNLOAD_TARGET_HELP = "Folder to download 'amiris-core_<version>-jar-with-dependencies.jar' to (default: './')"
DOWNLOAD_FORCE_HELP = "Force download to overwrite existing AMIRIS and/or example files (default: False)"
DOWNLOAD_MODE_HELP = "Choose to download model and examples `all` (default), or `model` / `examples` only"

RUN_HELP = "Compile scenario, execute AMIRIS, and extract results"
RUN_JAR_HELP = "Path to 'amiris-core_<version>-jar-with-dependencies.jar (default: 'amiris*.jar' in current directory)"
RUN_JAR_DEFAULT = "amiris*.jar"
RUN_JAR_ERR_MISSING = "No jar-file found with pattern: '{}'"
RUN_JAR_ERR_TOO_MANY = "More than one jar-file found with pattern: '{}'"
RUN_SCENARIO_HELP = "Path to a scenario yaml-file"
RUN_OUTPUT_HELP = "Directory to write output to"
RUN_OUTPUT_OPTION_HELP = (
    "optional pass through of FAME-Io's output conversion options, see "
    "https://gitlab.com/fame-framework/fame-io/-/blob/main/README.md#read-fame-results"
)
RUN_NO_CHECK_HELP = "Skip checks for Java installation and correct version"

BATCH_HELP = "Batch mode to perform multiple runs each with scenario compilation, execution, and results extraction"
BATCH_SCENARIO_HELP = "Path to single or list of: scenario yaml-files or their enclosing directories"
BATCH_RECURSIVE_HELP = "Option to recursively search in provided Path for scenario (default: False)"
DEFAULT_PATTERN = "*.y*ml"
BATCH_PATTERN_HELP = f"Optional name pattern that scenario files searched for must match (default: '{DEFAULT_PATTERN}')"

COMPARE_HELP = "Compare if results of two AMIRIS runs and equivalent"
COMPARE_EXPECTED_HELP = "Path to folder with expected results"
COMPARE_TEST_HELP = "Path to folder with results to test"
COMPARE_IGNORE_HELP = "Optional list of file names to not be compared"
URL_LATEST_AMIRIS = "https://gitlab.com/dlr-ve/esy/amiris/amiris/-/jobs/artifacts/main/download?job=deploy:jdk11"


class GeneralOptions(Enum):
    """Specifies general options for workflow"""

    LOG = auto()
    LOGFILE = auto()


class Command(Enum):
    """Specifies command to execute"""

    RUN = auto()
    DOWNLOAD = auto()
    COMPARE = auto()
    BATCH = auto()


class CompareOptions(Enum):
    """Options for command `compare`"""

    EXPECTED = auto()
    TEST = auto()
    IGNORE = auto()


class DownloadOptions(Enum):
    """Options for command `download`"""

    URL = auto()
    TARGET = auto()
    FORCE = auto()
    MODE = auto()


class RunOptions(Enum):
    """Options for command `run`"""

    JAR = auto()
    SCENARIO = auto()
    OUTPUT = auto()
    OUTPUT_OPTIONS = auto()
    NO_CHECKS = auto()


class BatchOptions(Enum):
    """Options for command `batch`"""

    JAR = auto()
    SCENARIOS = auto()
    OUTPUT = auto()
    RECURSIVE = auto()
    PATTERN = auto()
    NO_CHECKS = auto()


Options = {
    Command.COMPARE: CompareOptions,
    Command.RUN: RunOptions,
    Command.DOWNLOAD: DownloadOptions,
    Command.BATCH: BatchOptions,
}


def arg_handling_run(input_args: Optional[List[str]] = None) -> Tuple[Command, Dict[Enum, Any]]:
    """
    Handles command line arguments for `amiris` and returns `command` and its options `args`
    Allows to set args from a string through input_args

    Raises:
        CommandLineError: if a custom error occurred during command-line handling
    """

    parent_parser = argparse.ArgumentParser(prog="amiris", description=AMIRIS_PARSER)
    parent_parser.add_argument("-lf", "--logfile", type=Path, required=False, help=AMIRIS_LOG_FILE_HELP)
    parent_parser.add_argument(
        "-l",
        "--log",
        default=LogLevels.WARN.name,
        choices=[level.name.lower() for level in LogLevels],
        help=AMIRIS_LOG_LEVEL_HELP,
    )
    subparsers = parent_parser.add_subparsers(dest="command", required=True, help=AMIRIS_COMMAND_HELP)

    download_parser = subparsers.add_parser("download", help=DOWNLOAD_HELP)
    add_download_arguments(download_parser)

    install_parser = subparsers.add_parser("install", help=INSTALL_DEPRECATION)
    add_download_arguments(install_parser)

    run_parser = subparsers.add_parser("run", help=RUN_HELP)
    run_parser.add_argument("--scenario", "-s", type=Path, required=True, help=RUN_SCENARIO_HELP)
    run_parser.add_argument("--jar", "-j", type=str, default=RUN_JAR_DEFAULT, help=RUN_JAR_HELP)
    run_parser.add_argument("--output", "-o", type=Path, default=Path("./result"), help=RUN_OUTPUT_HELP)
    run_parser.add_argument("--output-options", "-oo", type=str, default="", help=RUN_OUTPUT_OPTION_HELP)
    run_parser.add_argument("--no-checks", "-nc", action="store_true", default=False, help=RUN_NO_CHECK_HELP)

    batch_parser = subparsers.add_parser("batch", help=BATCH_HELP)
    batch_parser.add_argument("--scenarios", "-s", nargs="+", type=Path, required=True, help=BATCH_SCENARIO_HELP)
    batch_parser.add_argument("--jar", "-j", type=str, default=RUN_JAR_DEFAULT, help=RUN_JAR_HELP)
    batch_parser.add_argument("--output", "-o", type=Path, default=Path("./"), help=RUN_OUTPUT_HELP)
    batch_parser.add_argument("--recursive", "-r", default=False, action="store_true", help=BATCH_RECURSIVE_HELP)
    batch_parser.add_argument("--pattern", "-p", type=str, default=DEFAULT_PATTERN, help=BATCH_PATTERN_HELP)
    batch_parser.add_argument("--no-checks", "-nc", action="store_true", default=False, help=RUN_NO_CHECK_HELP)

    compare_parser = subparsers.add_parser("compare", help=COMPARE_HELP)
    compare_parser.add_argument("--expected", "-e", type=Path, required=True, help=COMPARE_EXPECTED_HELP)
    compare_parser.add_argument("--test", "-t", type=Path, required=True, help=COMPARE_TEST_HELP)
    compare_parser.add_argument("--ignore", "-i", required=False, help=COMPARE_IGNORE_HELP)

    args = vars(parent_parser.parse_args(input_args))
    command = get_command_from_string(args.pop("command"))
    if command in (Command.RUN, Command.BATCH):
        args["jar"] = find_single_executable_path(args["jar"])
    return command, enumify(command, args)


def add_download_arguments(parser: argparse.ArgumentParser) -> None:
    """Adds download arguments to given parser"""
    parser.add_argument("--url", "-u", default=URL_LATEST_AMIRIS, help=DOWNLOAD_URL_MODEL_HELP)
    parser.add_argument("--target", "-t", type=Path, default=Path("./"), help=DOWNLOAD_TARGET_HELP)
    parser.add_argument("--force", "-f", default=False, action="store_true", help=DOWNLOAD_FORCE_HELP)
    parser.add_argument(
        "--mode",
        "-m",
        type=str.lower,  # noqa
        choices=[mode.name.lower() for mode in DownloadMode],
        default=DownloadMode.ALL.name,
        help=DOWNLOAD_MODE_HELP,
    )


def get_command_from_string(command_str: str) -> Command:
    """Returns Command extracted from given string"""
    if command_str == "install":
        print(INSTALL_DEPRECATION)
        command_str = "download"
    return Command[command_str.upper()]


def find_single_executable_path(pattern: str) -> Path:
    """Searches for single file with given `pattern`; Raises an error if no file or more than one file match."""
    matching_files = sorted(Path(".").glob(pattern))
    if len(matching_files) == 0:
        raise CommandLineError(RUN_JAR_ERR_MISSING.format(pattern))
    if len(matching_files) > 1:
        raise CommandLineError(RUN_JAR_ERR_TOO_MANY.format(pattern))
    return Path(matching_files[0])


def enumify(command: Command, args: dict) -> Dict[Enum, Any]:
    """Matches `args` for given `command` to their respective Enum"""

    result = {}
    for option in GeneralOptions:
        result[option] = args.pop(option.name.lower())

    for option in Options[command]:
        result[option] = args.pop(option.name.lower())
    return result



# SPDX-FileCopyrightText: 2022 German Aerospace Center <amiris@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

import logging as log
from enum import Enum
from typing import NoReturn


class LogLevels(Enum):
    """Levels for Logging"""

    PRINT = 100
    CRITICAL = log.CRITICAL
    ERROR = log.ERROR
    WARN = log.WARNING
    INFO = log.INFO
    DEBUG = log.DEBUG


def log_and_raise_critical(message: str) -> NoReturn:
    """
    Raises and logs a critical error with given message

    Args:
        message: to be used in Exception and logging
    """
    log.critical(message)
    raise Exception(message)


def log_error_and_raise(exception: Exception) -> NoReturn:
    """
    Raises the given `exception` and logs an error with message derived from the exceptions message

    Args:
        exception: to be raised and whose message will be logged as error
    """
    log.error(str(exception))
    raise exception


def set_up_logger(log_level: str, log_file_name: str = None) -> None:
    """Uses existing logger or sets up logger"""
    if not log.getLogger().hasHandlers():
        _set_up_new_logger(log_level, log_file_name)
    log.addLevelName(LogLevels.PRINT.value, LogLevels.PRINT.name)


def _set_up_new_logger(log_level: str, log_file_name: str = None) -> None:
    """
    Sets up root logger which always writes to console and optionally to file

    Args:
        log_level: the logging level to apply
        log_file_name: optional - if provided, logs will also be written to file
    """
    log_level = LogLevels[log_level.upper()]
    if log_level is LogLevels.DEBUG:
        formatter_string = (
            "%(asctime)s.%(msecs)03d — %(levelname)s — %(module)s:%(funcName)s:%(lineno)d — %(message)s"  # noqa
        )
    else:
        formatter_string = "%(asctime)s — %(levelname)s — %(message)s"  # noqa

    log_formatter = log.Formatter(formatter_string, "%H:%M:%S")
    root_logger = log.getLogger()
    root_logger.setLevel(log_level.name)

    if log_file_name:
        file_handler = log.FileHandler(log_file_name, mode="w")
        file_handler.setFormatter(log_formatter)
        root_logger.addHandler(file_handler)

    console_handler = log.StreamHandler()
    console_handler.setFormatter(log_formatter)
    root_logger.addHandler(console_handler)


def log_and_print(message: str) -> None:
    """
    Logs given message with the highest priority, thus ensures it is printed to console.

    Args:
        message: to be logged and printed
    """
    log.log(LogLevels.PRINT.value, message)



# SPDX-FileCopyrightText: 2022 German Aerospace Center <amiris@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

import logging
from pathlib import Path
from typing import List, Dict, Tuple

import numpy as np
import pandas as pd

_MISSING_EXPECTED_FILE = "Expected file {} was not found among files in the test folder."
_TEST_FILE_UNMATCHED = "Missing file to compare with file {} found in test folder."
_VALUE_DISALLOWED = "Found non-numeric and non-string value."
_ALL_VALUES_EQUIVALENT = "Content of files is equivalent for {}."
_VALUES_NOT_EQUIVALENT = "Files are not equivalent for {}."
_SHAPE_MISMATCH = "File to test {} has different shape than expected: {}"


def match_files(expected_files: List[Path], test_files: List[Path]) -> Dict[Path, Path]:
    """
    Compares files in given paths for equivalence

    Args:
        expected_files: paths of expected files
        test_files: path of files for text

    Returns:
        Match of each expected file to its test file, or None if test file is not found
    """
    expected_pairs = {item.stem.upper(): item for item in expected_files}
    test_pairs = {item.stem.upper(): item for item in test_files}
    result = {}
    for expected_stem, expected_file in expected_pairs.items():
        if expected_stem in test_pairs.keys():
            result[expected_file] = test_pairs.pop(expected_stem)
        else:
            result[expected_file] = ""
            logging.error(_MISSING_EXPECTED_FILE.format(expected_stem))
    if len(test_pairs) > 0:
        for test_stem in test_pairs.keys():
            logging.warning(_TEST_FILE_UNMATCHED.format(test_stem))
    return result


def compare_files(expected: Path, to_test: Path) -> str:
    """
    Compares two csv files for equivalence

    Args:
        expected: file with expected data
        to_test: file with data to test

    Raises:
        Exception: if given files are not equivalent

    Returns:
        Empty string if everything matched, else details on the differences
    """
    expected_df = read_file_and_sort_df(expected)
    test_df = read_file_and_sort_df(to_test)
    result = ""
    if expected_df.shape != test_df.shape:
        logging.error(_SHAPE_MISMATCH.format(to_test.stem, expected_df.shape))
        result = analyse_shape_difference(expected_df.shape, test_df.shape)
    else:
        all_values_compared = np.isclose(test_df, expected_df)
        if np.all(all_values_compared):
            logging.info(_ALL_VALUES_EQUIVALENT.format(to_test.stem))
        else:
            logging.error(_VALUES_NOT_EQUIVALENT.format(to_test.stem))
            result = analyse_row_difference(test_df, all_values_compared)
    return result


def analyse_shape_difference(expected_shape: Tuple[int, int], test_shape: Tuple[int, int]) -> str:
    """
    Analyses difference between two two-dimensional dataframe shapes

    Args:
        expected_shape: reference shape
        test_shape: shape to check against the reference

    Returns:
        description of the shape difference
    """
    row_delta = test_shape[0] - expected_shape[0]
    column_delta = test_shape[1] - expected_shape[1]
    result = ""
    if row_delta != 0:
        result += f"Test file {'has extra' if row_delta > 0 else 'misses' } {abs(row_delta)} row(s). "
    if column_delta != 0:
        result += f"Test file {'has extra' if row_delta > 0 else 'misses'} {abs(column_delta)} column(s). "
    return result


def analyse_row_difference(test_data: pd.DataFrame, comparison_result: np.ndarray) -> str:
    """
    Returns list of rows in the given test data frame that contain at least on mismatch in the comparison

    Args:
        test_data: original test data set with proper index
        comparison_result: array containing True or False for each row and column, where False indicates a mismatch to
                           the reference data set

    Returns:
        String listing all row numbers of the test data set that did not match the reference data
    """
    compare_df = pd.DataFrame(comparison_result, index=test_data.index)
    filtered = compare_df[(compare_df == False).any(axis=1)]  # noqa
    adjust_line_count_for_header_and_start_1 = [i + 2 for i in filtered.index.to_list()]
    return f"Deviations in test file line(s): {adjust_line_count_for_header_and_start_1}"


def read_file_and_sort_df(file: Path) -> pd.DataFrame:
    """
    Reads given file to data frame, replaces missing values with 0, converts strings to numbers
    and sorts from left to right columns

    Args:
        file: file to read

    Returns:
        sorted data frame with a numeric representation of given file's content (NaNs replaced by 0)
    """
    df = pd.read_csv(file, sep=";")
    df.fillna(0, inplace=True)
    df = df.apply(string_to_number)
    return df.sort_values(by=list(df.columns.values))


def string_to_number(series: pd.Series) -> pd.Series:
    """
    Converts pandas series of any numeric or string values to series of numeric values.
    Each string is converted to a single number: sum of its ascii values.
    Thus, identical strings always result in same number.
    However, different strings may also result in same number.
    Think of it as a very simple numeric hash function.
    Does not change numeric values in series.

    Args:
        series: series with numeric values and possibly mixed with strings

    Returns:
        Deterministic numeric series
    """
    return series.apply(lambda x: sum([ord(char) for char in x]) if isinstance(x, str) else x)



# SPDX-FileCopyrightText: 2023 German Aerospace Center <amiris@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

import logging as log
import re
import shutil
import subprocess

from amirispy.source.logs import log_and_raise_critical

_ERR_NO_JAVA = "No Java installation found. See {} for further instructions."
_ERR_JAVA_VERSION = "Local Java version '{}' does not match requirements '>{}'."
_URL_INSTALLATION_INSTRUCTIONS = "https://gitlab.com/dlr-ve/esy/amiris/amiris-py#further-requirements"

JAVA_VERSION_PATTERN = '"(\d+\.\d+).*"'  # noqa
JAVA_VERSION_MINIMUM = 11


def check_java_installation(raise_exception: bool = False) -> None:
    """If Java installation is not found, logs `Warning` (default) or raises Exception if `raise_exception`"""
    if not shutil.which("java"):
        if raise_exception:
            log_and_raise_critical(_ERR_NO_JAVA.format(_URL_INSTALLATION_INSTRUCTIONS))
        else:
            log.warning(_ERR_NO_JAVA.format(_URL_INSTALLATION_INSTRUCTIONS))


def check_java_version(raise_exception: bool = False) -> None:
    """If Java version is not compatible, logs `Warning` (default) or raises Exception if `raise_exception`"""
    version_raw = subprocess.check_output(["java", "-version"], stderr=subprocess.STDOUT)
    version_number = re.search(JAVA_VERSION_PATTERN, str(version_raw)).groups()[0]

    if float(version_number) < JAVA_VERSION_MINIMUM:
        if raise_exception:
            log_and_raise_critical(_ERR_JAVA_VERSION.format(version_number, JAVA_VERSION_MINIMUM))
        else:
            log.warning(_ERR_JAVA_VERSION.format(version_number, JAVA_VERSION_MINIMUM))


def check_java(skip: bool) -> None:
    """Checks java installation and version if not `skip`"""
    if not skip:
        check_java_installation(raise_exception=True)
        check_java_version(raise_exception=True)



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2022 German Aerospace Center <amiris@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

from typing import Optional, List

from amirispy.scripts.subcommands.batch import batch_run_amiris
from amirispy.scripts.subcommands.compare import compare_results
from amirispy.scripts.subcommands.download import download_amiris
from amirispy.scripts.subcommands.run import run_amiris
from amirispy.source.cli import (
    arg_handling_run,
    Command,
    CompareOptions,
    DownloadOptions,
    RunOptions,
    BatchOptions,
    GeneralOptions,
    CommandLineError,
)
from amirispy.source.logs import set_up_logger, log_and_print


def amiris_cli(args: Optional[List[str]] = None) -> None:
    """Calls sub-commands with appropriate arguments as returned by the command line parser"""
    try:
        command, options = arg_handling_run(args)
    except CommandLineError as e:
        raise SystemExit(e)

    set_up_logger(options[GeneralOptions.LOG], options[GeneralOptions.LOGFILE])
    if command is Command.DOWNLOAD:
        log_and_print("Starting install script")
        download_amiris(
            options[DownloadOptions.URL],
            options[DownloadOptions.TARGET],
            options[DownloadOptions.FORCE],
            options[DownloadOptions.MODE],
        )
        log_and_print(f"Installation setup to '{options[DownloadOptions.TARGET]}' complete")
    elif command is Command.RUN:
        log_and_print("Start running AMIRIS")
        run_amiris(options)
        log_and_print(f"Successfully executed AMIRIS. See your results in '{options[RunOptions.OUTPUT]}'")
    elif command is Command.BATCH:
        log_and_print("Start running AMIRIS batch run")
        batch_run_amiris(options)
        log_and_print(f"Successfully executed AMIRIS. See your results in '{options[BatchOptions.OUTPUT]}'")
    elif command is Command.COMPARE:
        log_and_print("Starting comparison script")
        compare_results(options[CompareOptions.EXPECTED], options[CompareOptions.TEST], options[CompareOptions.IGNORE])


if __name__ == "__main__":
    amiris_cli()



# SPDX-FileCopyrightText: 2022 German Aerospace Center <amiris@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

import logging as log
from pathlib import Path
from typing import List

from amirispy.source.file_comparison import match_files, compare_files
from amirispy.source.files import get_all_csv_files_in_folder_except
from amirispy.source.logs import log_and_print


def compare_results(folder_expected: Path, folder_to_test: Path, ignore_list: List[str] = None) -> None:
    """
    Compares content of two folders with AMIRIS results in CSV format for equivalence

    Args:
        folder_expected: folder with expected results
        folder_to_test: folder with results to test against expected results
        ignore_list: optional list of file names to ignore
    """

    expected_files = get_all_csv_files_in_folder_except(folder_expected, ignore_list)
    test_files = get_all_csv_files_in_folder_except(folder_to_test, ignore_list)
    file_pairs = match_files(expected_files, test_files)

    log.info(f"Checking {len(expected_files)} expected files...")
    results = {}
    for expected, to_test in file_pairs.items():
        if to_test:
            results[expected] = compare_files(expected, to_test)
        else:
            results[expected] = f"Missing file in test folder: {expected.stem}"

    differences_found = False
    for file, differences in results.items():
        if differences:
            differences_found = True
            log_and_print(f"FAIL: Found differences for {file.stem}: {differences}")

    if not differences_found:
        log_and_print("PASS: Found no significant differences for any expected pair of files.")



# SPDX-FileCopyrightText: 2022 German Aerospace Center <amiris@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0
import logging
import os
from pathlib import Path
from typing import List

from amirispy.source.logs import log_error_and_raise

_CSV_FILE_ENDING = ".csv"

_ERR_NOT_A_FOLDER = "Given Path '{}' is not a directory."
_ERR_MISSING_FOLDER = "Specified directory '{}' is missing."
_ERR_NO_ACCESS = "No writing permission to directory '{}'."

_WARN_NOT_EMPTY = "Folder '{}' is not empty - overriding files."


def get_all_csv_files_in_folder_except(folder: Path, exceptions: List[str] = None) -> List[Path]:
    """
    Find all csv files in a folder that can optionally ignore a files with a given file name

    Args:
        folder: to search for csv files - file ending is **not** case sensitive
        exceptions: optional, files names (without file ending) listed here will be ignored - **not** case sensitive

    Returns:
        Full file Paths for files ending with ".csv" not listed in exceptions
    """
    if not folder.is_dir():
        log_error_and_raise(ValueError(_ERR_MISSING_FOLDER.format(folder)))

    if exceptions is None:
        exceptions = list()
    exceptions = [item.upper() for item in exceptions]
    all_csvs = [file for file in folder.glob(f"*{_CSV_FILE_ENDING}")]
    return [file for file in all_csvs if file.stem not in exceptions]


def ensure_folder_exists(path: Path) -> None:
    """
    Returns Path to a directory and creates the folder if required.
    If given Path is an existing folder: does nothing, else creates new folder (including parent folders)

    Args:
        path: to check and create if not existing

    Returns:
        None

    Raises:
        ValueError: if path is an existing file
    """
    if path.is_file():
        log_error_and_raise(ValueError(_ERR_NOT_A_FOLDER.format(path)))
    if not path.is_dir():
        path.mkdir(parents=True)


def check_if_write_access(path: Path) -> None:
    """Raises Error if no writing access to `path`"""
    if not os.access(path, os.W_OK):
        log_error_and_raise(OSError(_ERR_NO_ACCESS.format(path)))


def warn_if_not_empty(folder: Path) -> None:
    """
    Logs a warning if given folder is not empty

    Args:
        folder: to check for files
    """
    if list(folder.glob("*")):
        logging.warning(_WARN_NOT_EMPTY.format(folder))



# SPDX-FileCopyrightText: 2024 German Aerospace Center <amiris@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

import logging as log
import os
from pathlib import Path
from typing import List

from fameio.input import SchemaError, ScenarioError
from fameio.input.loader import load_yaml
from fameio.input.scenario import Scenario
from fameio.input.validator import SchemaValidator, ValidationError

from amirispy.source.cli import BatchOptions
from amirispy.source.fameio_calls import determine_all_paths, compile_input, call_amiris, compile_output
from amirispy.source.files import check_if_write_access
from amirispy.source.logs import log_error_and_raise, log_and_print
from amirispy.source.util import check_java

_ERR_ALL_PATHS_INVALID = "Provided scenario path(s) '{}' contain no valid scenario YAML files."
_ERR_PATH_NOT_EXISTING = "Provided path '{}' is ignored as it is neither a file nor a directory."
_ERR_NO_VALID_FAME_SCENARIO = "'{}' is not a valid scenario file. You may improve the file name pattern using `-p`"


def batch_run_amiris(options: dict) -> None:
    """
    Compile multiple scenarios to protobuf using fameio.scripts.make_config, execute AMIRIS,
    and extract results using fameio.scripts.convert_results

    Args:
        options: dictionary of command line instructions
    """
    check_java(skip=options[BatchOptions.NO_CHECKS])
    origin_wd = Path.cwd()
    check_if_write_access(origin_wd)

    input_yaml_files = find_valid_scenarios(
        options[BatchOptions.SCENARIOS], options[BatchOptions.RECURSIVE], options[BatchOptions.PATTERN]
    )

    for i, input_yaml_file in enumerate(input_yaml_files):
        if len(input_yaml_files) >= 1:
            log_and_print(f"AMIRIS run {i+1}/{len(input_yaml_files)}")

        paths = determine_all_paths(input_yaml_file, origin_wd, options, batch=True)
        os.chdir(paths["SCENARIO_DIRECTORY"])
        compile_input(options, paths)
        os.chdir(origin_wd)
        call_amiris(paths)
        compile_output(options, paths)


def find_valid_scenarios(search_paths: List[Path], recursive: bool, pattern: str) -> List[Path]:
    """
    Searches for valid scenario YAML files in given `input_yaml_paths`

    Args:
        search_paths: path(s) which are to be searched for valid scenario files
        recursive: if true, subdirectories of each search path are searched as well
        pattern: that file names must match to be returned

    Returns:
        List of Paths to valid scenario files
    """
    files_to_test = get_inner_yaml_files(search_paths, recursive, pattern)
    scenario_files = [file for file in files_to_test if is_valid_fame_input_yaml(file)]
    if not scenario_files:
        log_error_and_raise(ValueError(_ERR_ALL_PATHS_INVALID.format(search_paths)))
    log.info(f"Found these scenario file(s) '{scenario_files}'.")
    return scenario_files


def get_inner_yaml_files(paths: List[Path], recursive: bool, pattern: str) -> List[Path]:
    """
    Returns a list of all YAML files in `paths` (and all subdirectories when `recursive` is set to True)

    Args:
        paths: to search for YAML files
        recursive: if True, subdirectories are also searched
        pattern: that file names must match to be returned

    Returns:
        List of Paths to YAML files contained in `paths` (and its subdirectories)
    """
    yaml_files = []
    for path in paths:
        if path.is_file():
            yaml_files.append(path)
        elif path.is_dir():
            yaml_files.extend([f for f in path.glob(pattern) if f.is_file()])
            if recursive:
                yaml_files.extend(get_inner_yaml_files([f for f in path.glob("*") if f.is_dir()], recursive, pattern))
        else:
            log.error(_ERR_PATH_NOT_EXISTING.format(path))
    return yaml_files


def is_valid_fame_input_yaml(file_to_test: Path) -> bool:
    """
    Checks if the given path points to a valid FAME input YAML file. If not, an error is logged.

    Args:
        file_to_test: Path to a file

    Returns:
        True if the given path points to valid FAME input YAML file, otherwise False
    """
    try:
        scenario = Scenario.from_dict(load_yaml(file_to_test))
        SchemaValidator.validate_scenario_and_timeseries(scenario)
        return True
    except (ScenarioError, SchemaError, FileNotFoundError, ValidationError) as e:
        log.warning(_ERR_NO_VALID_FAME_SCENARIO.format(file_to_test))
        log.info(f"Error: {e}")
    return False



# SPDX-FileCopyrightText: 2025 German Aerospace Center <amiris@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

import logging as log
import os
import shutil
import zipfile
from enum import Enum, auto
from pathlib import Path
from urllib.error import HTTPError

import wget

from amirispy.source.files import ensure_folder_exists, check_if_write_access
from amirispy.source.logs import log_and_print, log_and_raise_critical
from amirispy.source.util import check_java_installation, check_java_version

MSG_SKIPPED_DOWNLOAD = "Skipped download of {} due to option '-m/--mode {}'"
URL_EXAMPLES = "https://gitlab.com/dlr-ve/esy/amiris/examples/-/archive/main/examples-main.zip"

DEPLOY_URL = "https://gitlab.com/dlr-ve/esy/amiris/amiris/-/artifacts"
ERROR_URL_INVALID = "Download failed with error: '{}'. Please find latest deploy artifacts here: {}."


class DownloadMode(Enum):
    """Download `mode` options"""

    ALL = auto()
    MODEL = auto()
    EXAMPLES = auto()


def download_amiris(url: str, target_folder: Path, force: bool, mode: str) -> None:
    """
    Download and unzip AMIRIS from given url, overwriting existing AMIRIS file if `force` is enabled

    Args:
        url: where to download packaged AMIRIS file from
        target_folder: folder where to save the AMIRIS to
        force: flag to overwrite existing AMIRIS downloads of same version and existing examples
        mode: mode of download (all, model only, examples only)
    """

    ensure_folder_exists(target_folder)
    check_if_write_access(target_folder)

    mode = DownloadMode[mode.upper()]
    if mode == DownloadMode.ALL or mode == DownloadMode.MODEL:
        _conduct_model_download(url, target_folder, force)
    else:
        log.info(MSG_SKIPPED_DOWNLOAD.format("AMIRIS model", mode.name.lower()))

    if mode == DownloadMode.ALL or mode == DownloadMode.EXAMPLES:
        _conduct_example_download(target_folder, force)
    else:
        log.info(MSG_SKIPPED_DOWNLOAD.format("examples", mode.name.lower()))

    check_java_installation()
    check_java_version()


def _conduct_model_download(url: str, target_folder: Path, force: bool) -> None:
    """
    Downloads and extracts AMIRIS from `url` to `target_folder`. Overwrites existing files if `force` is True

    Args:
        url: where to download packaged AMIRIS file from
        target_folder: folder where to save the AMIRIS to
        force: flag to overwrite existing AMIRIS files
    """
    download_file_path = Path(target_folder, "amiris.zip")
    log.info("Starting download of AMIRIS")
    try:
        wget.download(url=url, out=str(download_file_path), bar=progress_bar)
    except HTTPError as e:
        log_and_raise_critical(ERROR_URL_INVALID.format(e, DEPLOY_URL))
    log.info(f"Downloaded file to '{download_file_path}'")

    if zipfile.is_zipfile(download_file_path):
        with zipfile.ZipFile(download_file_path, "r") as zip_ref:
            zip_ref.extractall(target_folder)
        os.remove(download_file_path)

        if force:
            for file in target_folder.glob("target/*with-dependencies.jar"):
                shutil.copy(src=str(file), dst=target_folder)
            log.info(f"Unzipped file content to '{target_folder}'")
        else:
            for file in target_folder.glob("target/*with-dependencies.jar"):
                try:
                    shutil.move(src=str(file), dst=target_folder)
                except shutil.Error:
                    log.error(
                        f"'{file.name}' already exists in '{target_folder}'. Use `-f/--force` to override anyway."
                    )
        shutil.rmtree(Path(target_folder, "target"))

    else:
        log.info("Downloaded file is not a zip file: Could not unzip.")
    log.info(f"Copying standard configuration files to '{target_folder}'.")
    log_and_print(f"AMIRIS download to '{target_folder}' completed.")


def progress_bar(current, total, _) -> str:
    """Progress bar that adds a newline in the end"""
    progress_message = "Download status: "
    if total < 0:
        progress_message += "unknown"
    elif current < total:
        progress_message += "%d%% [%d / %d] bytes" % (current / total * 100, current, total)
    else:
        progress_message += "done\n"
    return progress_message


def _conduct_example_download(target_folder: Path, force: bool) -> None:
    """
    Downloads and extracts examples for AMIRIS to `target_folder`. Overwrites existing examples if `force` is True

    Args:
        target_folder: folder where to save the examples to
        force: if True, existing examples are overwritten
    """
    download_file_path = Path(target_folder, "examples.zip")
    log.info("Starting download of examples")
    wget.download(url=URL_EXAMPLES, out=str(download_file_path), bar=progress_bar)
    print("")  # fix broken progress bar that misses the newline
    log.info(f"Downloaded examples to '{download_file_path}'")

    if Path(target_folder, "examples").exists() and not force:
        log.error(f"'examples' already exists in '{target_folder}'. Use `-f/--force` to override anyway.")
    else:
        if zipfile.is_zipfile(download_file_path):
            with zipfile.ZipFile(download_file_path, "r") as zip_ref:
                zip_ref.extractall(target_folder)
        else:
            log.info("Downloaded file is not a zip file: Could not unzip")
        shutil.move(src=f"{target_folder}/examples-main", dst=f"{target_folder}/examples")

        log_and_print(f"Examples download to '{target_folder}/examples' completed.")
    os.remove(download_file_path)



# SPDX-FileCopyrightText: 2024 German Aerospace Center <amiris@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

import logging as log
import subprocess
from pathlib import Path
from typing import Dict

from _ctypes import ArgumentError
from fameio.scripts.convert_results import run as convert_results
from fameio.scripts.make_config import run as make_config
from fameio.cli.convert_results import handle_args as handle_output_args
from fameio.cli.options import Options, ResolveOptions, TimeOptions

from amirispy.source.cli import BatchOptions, GeneralOptions, RunOptions
from amirispy.source.files import ensure_folder_exists, check_if_write_access, warn_if_not_empty
from amirispy.source.logs import log_error_and_raise

ERR_FAMEIO_OUTPUT_ARGS = "Invalid fameio output conversion options: '{}'"


def determine_all_paths(
    scenario_yaml: Path, working_directory: Path, options: Dict, batch: bool = False
) -> Dict[str, Path]:
    """
    Determines all Paths for the given scenario file and working directory

    Args:
        scenario_yaml: path to the scenario file
        working_directory: the base working directory
        options: amiris-py command line options
        batch: bool which is True when to determine paths for batch runs (default: False)

    Returns:
        Dictionary containing all Paths to final results and temporary files
    """
    paths = {
        "SCENARIO_FILE": scenario_yaml if scenario_yaml.is_absolute() else working_directory.joinpath(scenario_yaml),
        "SCENARIO_DIRECTORY": scenario_yaml.parent,
        "BASE_NAME": scenario_yaml.stem,
        "INPUT_PB": working_directory.joinpath("input.pb"),
        "OUTPUT_PB": working_directory.joinpath("output.pb"),
    }

    jar_option = BatchOptions.JAR if batch else RunOptions.JAR
    output_option = BatchOptions.OUTPUT if batch else RunOptions.OUTPUT
    paths.update(
        {
            "JAR_FILE": options[jar_option],
            "SETUP_FILE": Path(options[jar_option].parents[0], "fameSetup.yaml"),
            "RESULT_FOLDER": Path(options[output_option]),
        }
    )

    return paths


def compile_input(options: dict, paths: Dict[str, Path]) -> None:
    """
    Creates protobuf file using given `options` and `paths`

    Args:
        options: for logging
        paths: to given input and output files
    """
    fameio_input_config = {
        Options.FILE: paths["SCENARIO_FILE"].resolve(),
        Options.LOG_LEVEL: options[GeneralOptions.LOG],
        Options.LOG_FILE: options[GeneralOptions.LOGFILE],
        Options.OUTPUT: paths["INPUT_PB"],
    }
    log.info("Creating binary protobuf input file: `{protobuf_file_path}` from scenario `{scenario_file}`")
    make_config(fameio_input_config)


def call_amiris(paths: Dict[str, Path]) -> None:
    """
    Run AMIRIS using given paths

    Args:
        paths: to all required files
    """
    call = 'java -jar "{}" -f "{}" -o "output.pb"'.format(paths["JAR_FILE"], paths["INPUT_PB"])
    log.info(f"Calling AMIRIS with {paths['SCENARIO_FILE']}")
    subprocess.run(call, shell=True, check=True)


def compile_output(options: Dict, paths: Dict[str, Path]) -> None:
    """
    Conducts to conversion of AMIRIS output to CSV files based on given options and paths

    Args:
        options: for logging
        paths: to given input and output files
    """
    default_fameio_output_config = {
        Options.FILE: paths["OUTPUT_PB"].resolve(),
        Options.LOG_LEVEL: options[GeneralOptions.LOG],
        Options.LOG_FILE: options[GeneralOptions.LOGFILE],
        Options.AGENT_LIST: None,
        Options.OUTPUT: paths["RESULT_FOLDER"],
        Options.SINGLE_AGENT_EXPORT: False,
        Options.MEMORY_SAVING: False,
        Options.RESOLVE_COMPLEX_FIELD: ResolveOptions.SPLIT.name,
        Options.TIME: TimeOptions.UTC.name,
        Options.TIME_MERGING: [0, 1800, 1799],
    }

    ensure_folder_exists(paths["RESULT_FOLDER"])
    check_if_write_access(paths["RESULT_FOLDER"])
    log.info("Converting protobuf to csv files")
    warn_if_not_empty(paths["RESULT_FOLDER"])
    fameio_output_config = override_eligible_defaults(default_fameio_output_config, options[RunOptions.OUTPUT_OPTIONS])
    convert_results(fameio_output_config)


def override_eligible_defaults(defaults: Dict, overrides: str) -> Dict:
    """
    Replaces default fameio arguments with given overrides - ensures that the output folder and file are not modified

    Args:
        defaults: standard fameio output options
        overrides: (possibly empty) list of fameio output arguments to be used instead

    Returns:
        dict of (modified) fameio arguments
    """
    fixed_output_folder = defaults[Options.OUTPUT]
    fixed_file_to_convert = defaults[Options.FILE]
    try:
        argument_list = overrides.split()
        fameio_output_config = handle_output_args(argument_list, defaults)
    except SystemExit:
        log_error_and_raise(ArgumentError(ERR_FAMEIO_OUTPUT_ARGS.format(overrides)))
    fameio_output_config[Options.OUTPUT] = fixed_output_folder
    fameio_output_config[Options.FILE] = fixed_file_to_convert
    return fameio_output_config
